`timescale 1 ns/10 ps
/*
       opcode    functional_code 

DADD   000000    101100
DSUB   000000    101110
DAND   000000    100100
DOR    000000    101000
DXOR   000000    111100
SLT    000000    101010
DSRA   000000    111111
DSRL   000000    111010
DSLL   000000    111000


SLTI    001010    -----
LD      110111    ------
SD      111111    -------

*/
module CPU (clock);
parameter no_op = 32'b00000_100000;
input clock;
reg[63:0] PC, regbank[0:31];
reg [7:0] imem [0:255]; 
reg [7:0] dmem [0:255]; 
reg [63:0] IDEXA, IDEXB,  EXMEMB, // pipeline registers
           EXMEMALUOut, MEMWBValue ; // pipeline registers
reg [31:0] IFIDIR, IDEXIR, EXMEMIR, MEMWBIR;
wire [4:0] IDEXrs, IDEXrt, EXMEMrd, MEMWBrd, MEMWBrt; //hold register fields
wire [5:0] IDEXop,EXMEMop, MEMWBop; //Hold opcodes
wire stall;

initial $readmemb("instrfile1.dat", imem);
initial $readmemb("datafile1.dat", dmem);

assign IDEXrs = IDEXIR[25:21];
assign IDEXrt = IDEXIR[20:16];
assign EXMEMrd = EXMEMIR[15:11];
assign MEMWBrd = MEMWBIR[15:11];
assign EXMEMop = EXMEMIR[31:26];
assign MEMWBrt = MEMWBIR[20:16];
assign MEMWBop = MEMWBIR[31:26];
assign IDEXop =  IDEXIR[31:26];
assign stall = (MEMWBIR[31:26]==55) && // source instruction is a load
((((IDEXop==55)|(IDEXop==63)) && (IDEXrs==MEMWBrd)) | // stall for address calc
((IDEXop==ALUop) && ((IDEXrs==MEMWBrd)|(IDEXrt==MEMWBrd)))); // ALU use

reg [5:0] i; //used to initialize registers

initial begin
PC = 0;
IFIDIR = no_op; IDEXIR = no_op; EXMEMIR = no_op; MEMWBIR = no_op; // put no-ops in pipeline registers
for (i = 0;i<=31;i = i+1) 
begin
regbank[i] = i;

end
 //initialize registers--just so they aren’t cares
end
always @ (posedge clock) begin

if (~stall) begin // the first three pipeline stages stall if there is a load hazard
// first instruction in the pipeline is being fetched
// the IF stage
IFIDIR <= {imem [(PC>>2)], imem [(PC>>2)+1], imem [(PC>>2)+2], imem [(PC>>2)+3]};
PC <= PC + 4;
IDEXIR <= IFIDIR;
//pass along IR--can happen anywhere, since this affects next stage only!
// second instruction is in register fetch
// the ID stage
IDEXA <= regbank[IFIDIR[25:21]]; IDEXB <= regbank[IFIDIR[20:16]]; // get two registers
// third instruction is doing address calculation or ALU operation
// the EX stage
if ((IDEXop==55)|(IDEXop==63)) // address calculation & copy B
EXMEMALUOut <= IDEXA +{{48{IDEXIR[15]}}, IDEXIR[15:0]};
else if((IDEXop==10)) EXMEMALUOut <= (IDEXA < IDEXIR[15:0])? 1:0; 
else if (IDEXop==ALUop) case (IDEXIR[5:0]) //case for the various R-type instructions
44: EXMEMALUOut <= IDEXA + IDEXB; //add operation
46: EXMEMALUOut <= IDEXA - IDEXB; //sub operation
36: EXMEMALUOut <= IDEXA & IDEXB; //and operation
60: EXMEMALUOut <= IDEXA ^ IDEXB; //xor operation
40: EXMEMALUOut <= IDEXA | IDEXB; //or operation
42: EXMEMALUOut <= (IDEXA < IDEXB)?1:0 ; //slt operation
58: EXMEMALUOut <= IDEXB >> IDEXIR[10:6]; //dsrl operation
56: EXMEMALUOut <= IDEXB << IDEXIR[10:6]; //dsll operation
63: EXMEMALUOut <= IDEXB >> (IDEXIR[10:6]+64); //dsra operation
default: ; 
endcase
EXMEMIR <= IDEXIR; EXMEMB <= IDEXB; //pass along the IR & B register
end
else EXMEMIR <= no_op; //Freeze first three stages of pipeline; inject a nop into the EX output

$display("%b",IDEXA);
$display("%b",IDEXB);
$display("%b",EXMEMALUOut);

// the MEM stage
if (EXMEMop==ALUop) MEMWBValue <= EXMEMALUOut; //pass along ALU result

else if (EXMEMop == 55) MEMWBValue <= {dmem [(EXMEMALUOut>>2)], dmem [(EXMEMALUOut>>2)+1], dmem [(EXMEMALUOut>>2)+2],
dmem [(EXMEMALUOut>>2)+3],dmem [(EXMEMALUOut>>2)+4],dmem [(EXMEMALUOut>>2)+5],dmem [(EXMEMALUOut>>2)+6],dmem [(EXMEMALUOut>>2)+7]};

else if (EXMEMop == 63) {dmem[EXMEMALUOut>>2],dmem[(EXMEMALUOut>>2)+1],dmem[(EXMEMALUOut>>2)+2],dmem[(EXMEMALUOut>>2)+3]} <=EXMEMB; 

MEMWBIR <= EXMEMIR; //pass along IR

// the WB stage

if ((MEMWBop==ALUop) & (MEMWBrd != 0)) regbank[MEMWBrd] <= MEMWBValue; // ALU operation
else if ((EXMEMop == 63)& (MEMWBrt != 0)) regbank[MEMWBrt] <= MEMWBValue;
end

endmodule

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%% Test Bench %%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
module testbench();
reg clk;
CPU test(clk);
initial 
	begin	
		clk = 1'b1;
		forever 
		begin 
			#5 clk = ~clk;
			
		end
	end

initial begin
	#0
 	$display("Initiated test bench...\n\n");
	#150 $finish;
	$display("Ended test bench...\n\n");
end
initial begin
	$dumpfile("test2.vcd");
	$dumpvars(1,testbench.test);
end

endmodule

